@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
  * USA
  */
 
@@ -24,10 +24,17 @@
 #  include <config.h>
 #endif
 
-#include "pygobject-private.h"
-#include "pythread.h"
-#include <pyglib.h>
 #include <gobject/gvaluecollector.h>
+#include <pyglib.h>
+#include <pythread.h>
+#include "pygobject-private.h"
+#include "pygboxed.h"
+#include "pygenum.h"
+#include "pygflags.h"
+#include "pyginterface.h"
+#include "pygparamspec.h"
+#include "pygpointer.h"
+#include "pygtype.h"
 
 #ifdef HAVE_FFI_H
 #include "ffi-marshaller.h"
@@ -40,21 +47,6 @@ static PyObject *_pyg_signal_accumulator_true_handled_func;
 static GHashTable *log_handlers = NULL;
 static gboolean log_handlers_disabled = FALSE;
 
-GQuark pyginterface_type_key;
-GQuark pygobject_class_init_key;
-GQuark pyginterface_info_key;
-GQuark pygboxed_type_key;
-GQuark pygobject_class_key;
-GQuark pygobject_wrapper_key;
-GQuark pygboxed_marshal_key;
-GQuark pygenum_class_key;
-GQuark pygflags_class_key;
-GQuark pygpointer_class_key;
-GQuark pygobject_has_updated_constructor_key;
-GQuark pygobject_instance_data_key;
-
-
-
 static void pyg_flags_add_constants(PyObject *module, GType flags_type,
 				    const gchar *strip_prefix);
 
@@ -83,12 +75,6 @@ pyg_set_thread_block_funcs (PyGThreadBlockFunc block_threads_func,
 				  unblock_threads_func);
 }
 
-static void
-object_free(PyObject *op)
-{
-    PyObject_FREE(op);
-}
-
 /**
  * pyg_destroy_notify:
  * @user_data: a PyObject pointer.
@@ -108,141 +94,6 @@ pyg_destroy_notify(gpointer user_data)
 }
 
 
-/* ---------------- GBoxed functions -------------------- */
-
-GType PY_TYPE_OBJECT = 0;
-
-static gpointer
-pyobject_copy(gpointer boxed)
-{
-    PyObject *object = boxed;
-
-    Py_INCREF(object);
-    return object;
-}
-
-static void
-pyobject_free(gpointer boxed)
-{
-    PyObject *object = boxed;
-    PyGILState_STATE state;
-
-    state = pyglib_gil_state_ensure();
-    Py_DECREF(object);
-    pyglib_gil_state_release(state);
-}
-
-
-/* ---------------- GInterface functions -------------------- */
-
-static int
-pyg_interface_init(PyObject *self, PyObject *args, PyObject *kwargs)
-{
-    gchar buf[512];
-
-    if (!PyArg_ParseTuple(args, ":GInterface.__init__"))
-	return -1;
-
-    g_snprintf(buf, sizeof(buf), "%s can not be constructed", self->ob_type->tp_name);
-    PyErr_SetString(PyExc_NotImplementedError, buf);
-    return -1;
-}
-
-PyTypeObject PyGInterface_Type = {
-    PyObject_HEAD_INIT(NULL)
-    0,                                  /* ob_size */
-    "gobject.GInterface",               /* tp_name */
-    sizeof(PyObject),                   /* tp_basicsize */
-    0,                                  /* tp_itemsize */
-    /* methods */
-    (destructor)0,                      /* tp_dealloc */
-    (printfunc)0,                       /* tp_print */
-    (getattrfunc)0,                     /* tp_getattr */
-    (setattrfunc)0,                     /* tp_setattr */
-    (cmpfunc)0,                         /* tp_compare */
-    (reprfunc)0,                        /* tp_repr */
-    0,                                  /* tp_as_number */
-    0,                                  /* tp_as_sequence */
-    0,                                  /* tp_as_mapping */
-    (hashfunc)0,                        /* tp_hash */
-    (ternaryfunc)0,                     /* tp_call */
-    (reprfunc)0,                        /* tp_str */
-    (getattrofunc)0,                    /* tp_getattro */
-    (setattrofunc)0,                    /* tp_setattro */
-    0,					/* tp_as_buffer */
-    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,	/* tp_flags */
-    NULL, /* Documentation string */
-    (traverseproc)0,			/* tp_traverse */
-    (inquiry)0,				/* tp_clear */
-    (richcmpfunc)0,			/* tp_richcompare */
-    0,					/* tp_weaklistoffset */
-    (getiterfunc)0,			/* tp_iter */
-    (iternextfunc)0,			/* tp_iternext */
-    0,					/* tp_methods */
-    0,					/* tp_members */
-    0,					/* tp_getset */
-    (PyTypeObject *)0,			/* tp_base */
-    (PyObject *)0,			/* tp_dict */
-    0,					/* tp_descr_get */
-    0,					/* tp_descr_set */
-    0,					/* tp_dictoffset */
-    (initproc)pyg_interface_init,	/* tp_init */
-    (allocfunc)0,			/* tp_alloc */
-    (newfunc)0,				/* tp_new */
-    (freefunc)object_free,		/* tp_free */
-    (inquiry)0,				/* tp_is_gc */
-    (PyObject *)0,			/* tp_bases */
-};
-
-/**
- * pyg_register_interface:
- * @dict: a module dictionary.
- * @class_name: the class name for the wrapper class.
- * @gtype: the GType of the interface.
- * @type: the wrapper class for the interface.
- *
- * Registers a Python class as the wrapper for a GInterface.  As a
- * convenience it will also place a reference to the wrapper class in
- * the provided module dictionary.
- */
-static void
-pyg_register_interface(PyObject *dict, const gchar *class_name,
-                       GType gtype, PyTypeObject *type)
-{
-    PyObject *o;
-
-    type->ob_type = &PyType_Type;
-    type->tp_base = &PyGInterface_Type;
-
-    if (PyType_Ready(type) < 0) {
-        g_warning("could not ready `%s'", type->tp_name);
-        return;
-    }
-
-    if (gtype) {
-        o = pyg_type_wrapper_new(gtype);
-        PyDict_SetItemString(type->tp_dict, "__gtype__", o);
-        Py_DECREF(o);
-    }
-
-    g_type_set_qdata(gtype, pyginterface_type_key, type);
-    
-    PyDict_SetItemString(dict, (char *)class_name, (PyObject *)type);
-    
-}
-
-static void
-pyg_register_interface_info(GType gtype, const GInterfaceInfo *info)
-{
-    g_type_set_qdata(gtype, pyginterface_info_key, (gpointer) info);
-}
-
-static const GInterfaceInfo *
-pyg_lookup_interface_info(GType gtype)
-{
-    return g_type_get_qdata(gtype, pyginterface_info_key);
-}
-
 /* ---------------- gobject module functions -------------------- */
 
 static PyObject *
@@ -265,7 +116,7 @@ pyg_type_name (PyObject *self, PyObject *args)
 	return NULL;
     name = g_type_name(type);
     if (name)
-	return PyString_FromString(name);
+	return _PyUnicode_FromString(name);
     PyErr_SetString(PyExc_RuntimeError, "unknown typecode");
     return NULL;
 }
@@ -287,7 +138,7 @@ pyg_type_from_name (PyObject *self, PyObject *args)
     if (type != 0)
 	return pyg_type_wrapper_new(type);
     PyErr_Format(PyExc_RuntimeError, "%s: unknown type name: %s",
-		 PyString_AsString(PyObject_Repr((PyObject*)self)),
+		 _PyUnicode_AsString(PyObject_Repr((PyObject*)self)),
 		 name);
     return NULL;
 }
@@ -487,7 +338,7 @@ _pyg_signal_accumulator(GSignalInvocationHint *ihint,
 
     state = pyglib_gil_state_ensure();
     if (ihint->detail)
-        py_detail = PyString_FromString(g_quark_to_string(ihint->detail));
+        py_detail = _PyUnicode_FromString(g_quark_to_string(ihint->detail));
     else {
         Py_INCREF(Py_None);
         py_detail = Py_None;
@@ -646,17 +497,17 @@ add_signals (GType instance_type, PyObject *signals)
 	const gchar *signal_name;
         gchar *signal_name_canon, *c;
 
-	if (!PyString_Check(key)) {
+	if (!_PyUnicode_Check(key)) {
 	    PyErr_SetString(PyExc_TypeError,
 			    "__gsignals__ keys must be strings");
 	    ret = FALSE;
 	    break;
 	}
-	signal_name = PyString_AsString (key);
+	signal_name = _PyUnicode_AsString (key);
 
 	if (value == Py_None ||
-	    (PyString_Check(value) &&
-	     !strcmp(PyString_AsString(value), "override")))
+	    (_PyUnicode_Check(value) &&
+	     !strcmp(_PyUnicode_AsString(value), "override")))
         {
               /* canonicalize signal name, replacing '-' with '_' */
             signal_name_canon = g_strdup(signal_name);
@@ -933,7 +784,7 @@ pyg_param_spec_from_object (PyObject *tuple)
     }
     
     item = PyTuple_GetItem(tuple, val_length-1);
-    if (!PyInt_Check(item)) {
+    if (!_PyLong_Check(item)) {
 	PyErr_SetString(PyExc_TypeError,
 			"last element in tuple must be an int");
 	return NULL;
@@ -943,7 +794,7 @@ pyg_param_spec_from_object (PyObject *tuple)
     slice = PySequence_GetSlice(tuple, 4, val_length-1);
     pspec = create_property(prop_name, prop_type,
 			    nick, blurb, slice,
-			    PyInt_AsLong(item));
+			    _PyLong_AsLong(item));
 
     return pspec;
 }
@@ -968,13 +819,13 @@ add_properties (GType instance_type, PyObject *properties)
 	
 	/* values are of format (type,nick,blurb, type_specific_args, flags) */
 	
-	if (!PyString_Check(key)) {
+	if (!_PyUnicode_Check(key)) {
 	    PyErr_SetString(PyExc_TypeError,
 			    "__gproperties__ keys must be strings");
 	    ret = FALSE;
 	    break;
 	}
-	prop_name = PyString_AsString (key);
+	prop_name = _PyUnicode_AsString (key);
 
 	if (!PyTuple_Check(value)) {
 	    PyErr_SetString(PyExc_TypeError,
@@ -1007,13 +858,13 @@ add_properties (GType instance_type, PyObject *properties)
 	    break;
 	}
 	item = PyTuple_GetItem(value, val_length-1);
-	if (!PyInt_Check(item)) {
+	if (!_PyLong_Check(item)) {
 	    PyErr_SetString(PyExc_TypeError,
 		"last element in __gproperties__ value tuple must be an int");
 	    ret = FALSE;
 	    break;
 	}
-	flags = PyInt_AsLong(item);
+	flags = _PyLong_AsLong(item);
 
 	/* slice is the extra items in the tuple */
 	slice = PySequence_GetSlice(value, 3, val_length-1);
@@ -1027,14 +878,14 @@ add_properties (GType instance_type, PyObject *properties)
             PyObject *type, *value, *traceback;
 	    ret = FALSE;
             PyErr_Fetch(&type, &value, &traceback);
-            if (PyString_Check(value)) {
+            if (_PyUnicode_Check(value)) {
                 char msg[256];
                 g_snprintf(msg, 256,
 			   "%s (while registering property '%s' for GType '%s')",
-                           PyString_AsString(value),
+                           _PyUnicode_AsString(value),
 			   prop_name, g_type_name(instance_type));
                 Py_DECREF(value);
-                value = PyString_FromString(msg);
+                value = _PyUnicode_FromString(msg);
             }
             PyErr_Restore(type, value, traceback);
 	    break;
@@ -1114,17 +965,18 @@ get_type_name_for_class(PyTypeObject *class)
     gint i, name_serial;
     char name_serial_str[16];
     PyObject *module;
-    char *type_name;
+    char *type_name = NULL;
     
     /* make name for new GType */
     name_serial = 1;
     /* give up after 1000 tries, just in case.. */
     while (name_serial < 1000) 
     {
+	g_free(type_name);
 	snprintf(name_serial_str, 16, "-v%i", name_serial);
 	module = PyObject_GetAttrString((PyObject *)class, "__module__");
-	if (module && PyString_Check(module)) {
-	    type_name = g_strconcat(PyString_AsString(module), ".",
+	if (module && _PyUnicode_Check(module)) {
+	    type_name = g_strconcat(_PyUnicode_AsString(module), ".",
 				    class->tp_name,
 				    name_serial > 1 ? name_serial_str : NULL,
 				    NULL);
@@ -1194,7 +1046,7 @@ pygobject__g_instance_init(GTypeInstance   *instance,
         wrapper = pygobject_new_full(object, FALSE, g_class);
         args = PyTuple_New(0);
         kwargs = PyDict_New();
-        if (wrapper->ob_type->tp_init(wrapper, args, kwargs))
+        if (Py_TYPE(wrapper)->tp_init(wrapper, args, kwargs))
             PyErr_Print();
         Py_DECREF(args);
         Py_DECREF(kwargs);
@@ -1202,6 +1054,77 @@ pygobject__g_instance_init(GTypeInstance   *instance,
     }
 }
 
+
+/*  This implementation is bad, see bug 566571 for an example why.
+ *  Instead of scanning explicitly declared bases for interfaces, we
+ *  should automatically initialize all implemented interfaces to
+ *  prevent bugs like that one.  However, this will lead to
+ *  performance degradation as each virtual method in derived classes
+ *  will round-trip through do_*() stuff, *even* if it is not
+ *  overriden.  We need to teach codegen to retain parent method
+ *  instead of setting virtual to *_proxy_do_*() if corresponding
+ *  do_*() is not overriden.  Ok, that was a messy explanation.
+ */
+static void
+pyg_type_add_interfaces(PyTypeObject *class, GType instance_type,
+                        PyObject *bases, gboolean new_interfaces,
+                        GType *parent_interfaces, guint n_parent_interfaces)
+{
+    int i;
+
+    if (!bases) {
+        g_warning("type has no bases");
+        return;
+    }
+
+    for (i = 0; i < PyTuple_GET_SIZE(bases); ++i) {
+        guint k;
+        PyObject *base = PyTuple_GET_ITEM(bases, i);
+        GType itype;
+        gboolean is_new = TRUE;
+        const GInterfaceInfo *iinfo;
+        GInterfaceInfo iinfo_copy;
+
+        /* 'base' can also be a PyClassObject, see bug #566571. */
+        if (!PyType_Check(base))
+            continue;
+
+        if (!PyType_IsSubtype((PyTypeObject*) base, &PyGInterface_Type))
+            continue;
+
+        itype = pyg_type_from_object(base);
+
+        /* Happens for _implementations_ of an interface. */
+        if (!G_TYPE_IS_INTERFACE(itype))
+            continue;
+
+        for (k = 0; k < n_parent_interfaces; ++k) {
+            if (parent_interfaces[k] == itype) {
+                is_new = FALSE;
+                break;
+            }
+        }
+
+        if ((new_interfaces && !is_new) || (!new_interfaces && is_new))
+            continue;
+
+        iinfo = pyg_lookup_interface_info(itype);
+        if (!iinfo) {
+            gchar *error;
+            error = g_strdup_printf("Interface type %s "
+                                    "has no Python implementation support",
+                                    ((PyTypeObject *) base)->tp_name);
+            PyErr_Warn(PyExc_RuntimeWarning, error);
+            g_free(error);
+            continue;
+        }
+
+        iinfo_copy = *iinfo;
+        iinfo_copy.interface_data = class;
+        g_type_add_interface_static(instance_type, itype, &iinfo_copy);
+    }
+}
+
 int
 pyg_type_register(PyTypeObject *class, const char *type_name)
 {
@@ -1209,7 +1132,6 @@ pyg_type_register(PyTypeObject *class, const char *type_name)
     GType parent_type, instance_type;
     GType *parent_interfaces;
     guint n_parent_interfaces;
-    gint i;
     GTypeQuery query;
     gpointer gclass;
     gpointer has_new_constructor_api;
@@ -1231,9 +1153,8 @@ pyg_type_register(PyTypeObject *class, const char *type_name)
 
     /* find the GType of the parent */
     parent_type = pyg_type_from_object((PyObject *)class);
-    if (!parent_type) {
+    if (!parent_type)
 	return -1;
-    }
 
     parent_interfaces = g_type_interfaces(parent_type, &n_parent_interfaces);
 
@@ -1288,53 +1209,37 @@ pyg_type_register(PyTypeObject *class, const char *type_name)
 			     pyg_object_descr_doc_get());
     }
 
-      /* Register interfaces that are already defined by the parent
-       * type and are going to be reimplemented  */
-    if (class->tp_bases) {
-        for (i = 0; i < PyTuple_GET_SIZE(class->tp_bases); ++i)
-        {
-            PyTypeObject *base =
-		(PyTypeObject *) PyTuple_GET_ITEM(class->tp_bases, i);
-            GType itype;
-            const GInterfaceInfo *iinfo;
-            GInterfaceInfo iinfo_copy;
-            guint parent_interface_iter;
-
-            if (((PyTypeObject *) base)->tp_base != &PyGInterface_Type)
-                continue;
-
-            itype = pyg_type_from_object((PyObject *) base);
-
-              /* ignore interface unless defined in parent type */
-            if (n_parent_interfaces == 0)
-                continue;
-            for (parent_interface_iter = 0;
-                 parent_interface_iter < n_parent_interfaces;
-                 ++parent_interface_iter)
-            {
-                if (parent_interfaces[parent_interface_iter] == itype)
-                    break;
-            }
-            if (parent_interface_iter == n_parent_interfaces)
-                continue;
-
-            iinfo = pyg_lookup_interface_info(itype);
-            if (!iinfo) {
-                char *msg;
-                msg = g_strdup_printf("Interface type %s "
-                                      "has no python implementation support",
-                                      base->tp_name);
-                PyErr_Warn(PyExc_RuntimeWarning, msg);
-                g_free(msg);
-                continue;
-            }
-
-            iinfo_copy = *iinfo;
-            iinfo_copy.interface_data = class;
-            g_type_add_interface_static(instance_type, itype, &iinfo_copy);
-        }
-    } else
-        g_warning("type has no tp_bases");
+    /*
+     * Note: Interfaces to be implemented are searched twice.  First
+     * we register interfaces that are already implemented by a parent
+     * type.  The second time, the remaining interfaces are
+     * registered, i.e. the ones that are not implemented by a parent
+     * type.  In between these two loops, properties and signals are
+     * registered.  It has to be done this way, in two steps,
+     * otherwise glib will complain.  If registering all interfaces
+     * always before properties, you get an error like:
+     *
+     *    ../gobject:121: Warning: Object class
+     *    test_interface+MyObject doesn't implement property
+     *    'some-property' from interface 'TestInterface'
+     *
+     * If, on the other hand, you register interfaces after
+     * registering the properties, you get something like:
+     *
+     *     ../gobject:121: Warning: cannot add interface type
+     *    `TestInterface' to type `test_interface+MyUnknown', since
+     *    type `test_interface+MyUnknown' already conforms to
+     *    interface
+     *
+     * This looks like a GLib quirk, but no bug has been filed
+     * upstream.  However we have a unit test for this particular
+     * problem, which can be found in test_interfaces.py, class
+     * TestInterfaceImpl.
+     *
+     * See also comment above pyg_type_add_interfaces().
+     */
+    pyg_type_add_interfaces(class, instance_type, class->tp_bases, FALSE,
+                            parent_interfaces, n_parent_interfaces);
 
     /* we look this up in the instance dictionary, so we don't
      * accidentally get a parent type's __gsignals__ attribute. */
@@ -1380,54 +1285,11 @@ pyg_type_register(PyTypeObject *class, const char *type_name)
 	PyErr_Clear();
     }
 
-      /* Register new interfaces, that are _not_ already defined by
-       * the parent type  */
-    if (class->tp_bases) {
-        for (i = 0; i < PyTuple_GET_SIZE(class->tp_bases); ++i)
-        {
-            PyTypeObject *base =
-		(PyTypeObject *) PyTuple_GET_ITEM(class->tp_bases, i);
-            GType itype;
-            const GInterfaceInfo *iinfo;
-            GInterfaceInfo iinfo_copy;
-            guint parent_interface_iter;
-
-            if (((PyTypeObject *) base)->tp_base != &PyGInterface_Type)
-                continue;
-
-            itype = pyg_type_from_object((PyObject *) base);
-
-              /* ignore interface if already defined in parent type */
-            if (n_parent_interfaces != 0) {
-                for (parent_interface_iter = 0;
-                     parent_interface_iter < n_parent_interfaces;
-                     ++parent_interface_iter)
-                {
-                    if (parent_interfaces[parent_interface_iter] == itype)
-                        break;
-                }
-
-                if (parent_interface_iter < n_parent_interfaces)
-                    continue;
-            }
-
-            iinfo = pyg_lookup_interface_info(itype);
-            if (!iinfo) {
-                char *msg;
-                msg = g_strdup_printf("Interface type %s "
-                                      "has no python implementation support",
-                                      base->tp_name);
-                PyErr_Warn(PyExc_RuntimeWarning, msg);
-                g_free(msg);
-                continue;
-            }
-
-            iinfo_copy = *iinfo;
-            iinfo_copy.interface_data = class;
-            g_type_add_interface_static(instance_type, itype, &iinfo_copy);
-        }
-    } else
-        g_warning("type has no tp_bases");
+    /* Register new interfaces, that are _not_ already defined by
+     * the parent type.  FIXME: See above.
+     */
+    pyg_type_add_interfaces(class, instance_type, class->tp_bases, TRUE,
+                            parent_interfaces, n_parent_interfaces);
 
     gclass = g_type_class_ref(instance_type);
     if (pyg_run_class_init(instance_type, gclass, class)) {
@@ -1463,12 +1325,20 @@ pyg_signal_new(PyObject *self, PyObject *args)
 			  &py_type, &signal_flags, &py_return_type,
 			  &py_param_types))
 	return NULL;
+
     instance_type = pyg_type_from_object(py_type);
     if (!instance_type)
 	return NULL;
+    if (!(G_TYPE_IS_INSTANTIATABLE(instance_type) || G_TYPE_IS_INTERFACE(instance_type))) {
+	PyErr_SetString(PyExc_TypeError,
+			"argument 2 must be an object type or interface type");
+	return NULL;
+    }
+    
     return_type = pyg_type_from_object(py_return_type);
     if (!return_type)
 	return NULL;
+
     if (!PySequence_Check(py_param_types)) {
 	PyErr_SetString(PyExc_TypeError,
 			"argument 5 must be a sequence of GType codes");
@@ -1498,7 +1368,7 @@ pyg_signal_new(PyObject *self, PyObject *args)
 			      return_type, n_params, param_types);
     g_free(param_types);
     if (signal_id != 0)
-	return PyInt_FromLong(signal_id);
+	return _PyLong_FromLong(signal_id);
     PyErr_SetString(PyExc_RuntimeError, "could not create signal");
     return NULL;
 }
@@ -1543,7 +1413,7 @@ pyg_signal_list_names (PyObject *self, PyObject *args, PyObject *kwargs)
     if (list != NULL) {
 	for (i = 0; i < n; i++)
 	    PyTuple_SetItem(list, i,
-			    PyString_FromString(g_signal_name(ids[i])));
+			    _PyUnicode_FromString(g_signal_name(ids[i])));
     }
     
     g_free(ids);
@@ -1599,7 +1469,7 @@ pyg_signal_list_ids (PyObject *self, PyObject *args, PyObject *kwargs)
     }
 
     for (i = 0; i < n; i++)
-	PyTuple_SetItem(list, i, PyInt_FromLong(ids[i]));
+	PyTuple_SetItem(list, i, _PyLong_FromLong(ids[i]));
     g_free(ids);
     if (class)
         g_type_class_unref(class);
@@ -1647,7 +1517,7 @@ pyg_signal_lookup (PyObject *self, PyObject *args, PyObject *kwargs)
         g_type_class_unref(class);
     else
        g_type_default_interface_unref(iface);
-    return PyInt_FromLong(id);
+    return _PyLong_FromLong(id);
 }
 
 static PyObject *
@@ -1662,7 +1532,7 @@ pyg_signal_name (PyObject *self, PyObject *args, PyObject *kwargs)
 	return NULL;
     signal_name = g_signal_name(id);
     if (signal_name)
-        return PyString_FromString(signal_name);
+        return _PyUnicode_FromString(signal_name);
 
     Py_INCREF(Py_None);
     return Py_None;
@@ -1735,10 +1605,10 @@ pyg_signal_query (PyObject *self, PyObject *args, PyObject *kwargs)
         goto done;
     }
 
-    PyTuple_SET_ITEM(py_query, 0, PyInt_FromLong(query.signal_id));
-    PyTuple_SET_ITEM(py_query, 1, PyString_FromString(query.signal_name));
+    PyTuple_SET_ITEM(py_query, 0, _PyLong_FromLong(query.signal_id));
+    PyTuple_SET_ITEM(py_query, 1, _PyUnicode_FromString(query.signal_name));
     PyTuple_SET_ITEM(py_query, 2, pyg_type_wrapper_new(query.itype));
-    PyTuple_SET_ITEM(py_query, 3, PyInt_FromLong(query.signal_flags));
+    PyTuple_SET_ITEM(py_query, 3, _PyLong_FromLong(query.signal_flags));
     PyTuple_SET_ITEM(py_query, 4, pyg_type_wrapper_new(query.return_type));
     for (i = 0; i < query.n_params; i++) {
         PyTuple_SET_ITEM(params_list, i,
@@ -1849,7 +1719,7 @@ pyg_object_new (PyGObject *self, PyObject *args, PyObject *kwargs)
 	params = g_new0(GParameter, PyDict_Size(kwargs));
 	while (PyDict_Next (kwargs, &pos, &key, &value)) {
 	    GParamSpec *pspec;
-	    const gchar *key_str = PyString_AsString (key);
+	    const gchar *key_str = _PyUnicode_AsString (key);
 
 	    pspec = g_object_class_find_property (class, key_str);
 	    if (!pspec) {
@@ -1864,7 +1734,7 @@ pyg_object_new (PyGObject *self, PyObject *args, PyObject *kwargs)
 					       value, pspec) < 0) {
 		PyErr_Format(PyExc_TypeError,
 			     "could not convert value for property `%s' from %s to %s",
-			     key_str, value->ob_type->tp_name,
+			     key_str, Py_TYPE(value)->tp_name,
 			     g_type_name(G_PARAM_SPEC_VALUE_TYPE(pspec)));
 		goto cleanup;
 	    }
@@ -1944,6 +1814,21 @@ pyg_threads_init (PyObject *unused, PyObject *args, PyObject *kwargs)
     return Py_None;
 }
 
+/* Only for backwards compatibility */
+int
+pygobject_enable_threads(void)
+{
+    if (!pyglib_enable_threads())
+	return -1;
+    return 0;
+}
+
+static void
+pyg_note_threads_enabled(void)
+{
+    pygobject_api_functions.threads_enabled = TRUE;
+}
+
 static PyObject *
 pyg_signal_accumulator_true_handled(PyObject *unused, PyObject *args)
 {
@@ -2037,7 +1922,7 @@ pyg_add_emission_hook(PyGObject *self, PyObject *args)
 
     if (!g_signal_parse_name(name, gtype, &sigid, &detail, TRUE)) {
 	PyErr_Format(PyExc_TypeError, "%s: unknown signal name: %s",
-		     PyString_AsString(PyObject_Repr((PyObject*)self)),
+		     _PyUnicode_AsString(PyObject_Repr((PyObject*)self)),
 		     name);
 	return NULL;
     }
@@ -2076,7 +1961,7 @@ pyg_remove_emission_hook(PyGObject *self, PyObject *args)
     
     if (!g_signal_parse_name(name, gtype, &signal_id, NULL, TRUE)) {
 	PyErr_Format(PyExc_TypeError, "%s: unknown signal name: %s",
-		     PyString_AsString(PyObject_Repr((PyObject*)self)),
+		     _PyUnicode_AsString(PyObject_Repr((PyObject*)self)),
 		     name);
 	return NULL;
     }
@@ -2093,12 +1978,14 @@ pyg__install_metaclass(PyObject *dummy, PyTypeObject *metaclass)
     Py_INCREF(metaclass);
     PyGObject_MetaType = metaclass;
     Py_INCREF(metaclass);
-    PyGObject_Type.ob_type = metaclass;
+
+    Py_TYPE(&PyGObject_Type) = metaclass;
+
     Py_INCREF(Py_None);
     return Py_None;
 }
 
-static PyMethodDef pygobject_functions[] = {
+static PyMethodDef _gobject_functions[] = {
     { "type_name", pyg_type_name, METH_VARARGS },
     { "type_from_name", pyg_type_from_name, METH_VARARGS },
     { "type_parent", pyg_type_parent, METH_VARARGS },
@@ -2305,44 +2192,6 @@ pyg_gerror_exception_check(GError **error)
     return pyglib_gerror_exception_check(error);
 }
 
-static PyObject *
-_pyg_strv_from_gvalue(const GValue *value)
-{
-    gchar    **argv = (gchar **) g_value_get_boxed(value);
-    int        argc = 0, i;
-    PyObject  *py_argv;
-
-    if (argv) {
-        while (argv[argc])
-            argc++;
-    }
-    py_argv = PyList_New(argc);
-    for (i = 0; i < argc; ++i)
-	PyList_SET_ITEM(py_argv, i, PyString_FromString(argv[i]));
-    return py_argv;
-}
-
-static int
-_pyg_strv_to_gvalue(GValue *value, PyObject *obj)
-{
-    Py_ssize_t argc, i;
-    gchar **argv;
-
-    if (!(PyTuple_Check(obj) || PyList_Check(obj)))
-        return -1;
-
-    argc = PySequence_Length(obj);
-    for (i = 0; i < argc; ++i)
-	if (!PyString_Check(PySequence_Fast_GET_ITEM(obj, i)))
-	    return -1;
-    argv = g_new(gchar *, argc + 1);
-    for (i = 0; i < argc; ++i)
-	argv[i] = g_strdup(PyString_AsString(PySequence_Fast_GET_ITEM(obj, i)));
-    argv[i] = NULL;
-    g_value_take_boxed(value, argv);
-    return 0;
-}
-
 /**
  * pyg_parse_constructor_args: helper function for PyGObject constructors
  * @obj_type: GType of the GObject, for parameter introspection
@@ -2493,7 +2342,6 @@ pyg_set_object_has_new_constructor(GType type)
     g_type_set_qdata(type, pygobject_has_updated_constructor_key, GINT_TO_POINTER(1));
 }
 
-#define GET_INT(x) (((PyIntObject*)x)->ob_ival)
 PyObject *
 pyg_integer_richcompare(PyObject *v, PyObject *w, int op)
 {
@@ -2501,12 +2349,12 @@ pyg_integer_richcompare(PyObject *v, PyObject *w, int op)
     gboolean t;
 
     switch (op) {
-    case Py_EQ: t = GET_INT(v) == GET_INT(w); break;
-    case Py_NE: t = GET_INT(v) != GET_INT(w); break;
-    case Py_LE: t = GET_INT(v) <= GET_INT(w); break;
-    case Py_GE: t = GET_INT(v) >= GET_INT(w); break;
-    case Py_LT: t = GET_INT(v) <  GET_INT(w); break;
-    case Py_GT: t = GET_INT(v) >  GET_INT(w); break;
+    case Py_EQ: t = _PyLong_AS_LONG(v) == _PyLong_AS_LONG(w); break;
+    case Py_NE: t = _PyLong_AS_LONG(v) != _PyLong_AS_LONG(w); break;
+    case Py_LE: t = _PyLong_AS_LONG(v) <= _PyLong_AS_LONG(w); break;
+    case Py_GE: t = _PyLong_AS_LONG(v) >= _PyLong_AS_LONG(w); break;
+    case Py_LT: t = _PyLong_AS_LONG(v) <  _PyLong_AS_LONG(w); break;
+    case Py_GT: t = _PyLong_AS_LONG(v) >  _PyLong_AS_LONG(w); break;
     default: g_assert_not_reached();
     }
 
@@ -2514,7 +2362,6 @@ pyg_integer_richcompare(PyObject *v, PyObject *w, int op)
     Py_INCREF(result);
     return result;
 }
-#undef GET_INT
 
 static void
 _log_func(const gchar *log_domain,
@@ -2639,7 +2486,7 @@ struct _PyGObject_Functions pygobject_api_functions = {
   pyg_flags_from_gtype,
 
   FALSE, /* threads_enabled */
-  pyglib_enable_threads,
+  pygobject_enable_threads,
   pygobject_gil_state_ensure,
   pygobject_gil_state_release,
   pyg_register_class_init,
@@ -2660,123 +2507,21 @@ struct _PyGObject_Functions pygobject_api_functions = {
   
 };
 
-#define REGISTER_TYPE(d, type, name) \
-    type.ob_type = &PyType_Type; \
-    if (!type.tp_alloc) \
-	type.tp_alloc = PyType_GenericAlloc; \
-    if (!type.tp_new) \
-	type.tp_new = PyType_GenericNew; \
-    if (PyType_Ready(&type)) \
-	return; \
-    PyDict_SetItemString(d, name, (PyObject *)&type);
-
-#define REGISTER_GTYPE(d, type, name, gtype) \
-    REGISTER_TYPE(d, type, name); \
-    PyDict_SetItemString(type.tp_dict, "__gtype__", \
-			 o=pyg_type_wrapper_new(gtype)); \
-    Py_DECREF(o);
-
-DL_EXPORT(void)
-init_gobject(void)
+/* for addon libraries ... */
+static void
+pygobject_register_api(PyObject *d)
 {
-    PyObject *m, *d, *o, *tuple, *features;
-    PyObject *descr;
-    PyObject *warning;
+    PyObject *api;
     
-    PyGParamSpec_Type.ob_type = &PyType_Type;
-    if (PyType_Ready(&PyGParamSpec_Type))
-	return;
-    
-    m = Py_InitModule("gobject._gobject", pygobject_functions);
-    d = PyModule_GetDict(m);
+    api = PyCObject_FromVoidPtr(&pygobject_api_functions,NULL);
+    PyDict_SetItemString(d, "_PyGObject_API", api);
+    Py_DECREF(api);
+}
 
-    PyDict_SetItemString(d, "GParamSpec", (PyObject *)&PyGParamSpec_Type);
-
-    g_type_init();
-    pyglib_init();
-
-    pygboxed_type_key        = g_quark_from_static_string("PyGBoxed::class");
-    pygboxed_marshal_key     = g_quark_from_static_string("PyGBoxed::marshal");
-    pygenum_class_key        = g_quark_from_static_string("PyGEnum::class");
-    pygflags_class_key       = g_quark_from_static_string("PyGFlags::class");
-    pygobject_class_key      = g_quark_from_static_string("PyGObject::class");
-    pygobject_class_init_key = g_quark_from_static_string("PyGObject::class-init");
-    pygobject_wrapper_key    = g_quark_from_static_string("PyGObject::wrapper");
-    pyginterface_type_key    = g_quark_from_static_string("PyGInterface::type");
-    pyginterface_info_key    = g_quark_from_static_string("PyGInterface::info");
-    pygpointer_class_key     = g_quark_from_static_string("PyGPointer::class");
-    pygobject_has_updated_constructor_key =\
-        g_quark_from_static_string("PyGObject::has-updated-constructor");
-    pygobject_instance_data_key = g_quark_from_static_string("PyGObject::instance-data");
-
-    REGISTER_TYPE(d, PyGTypeWrapper_Type, "GType");
-
-    if (!PY_TYPE_OBJECT)
-	PY_TYPE_OBJECT = g_boxed_type_register_static("PyObject",
-						      pyobject_copy,
-						      pyobject_free);
-
-    PyGObject_Type.tp_alloc = PyType_GenericAlloc;
-    PyGObject_Type.tp_new = PyType_GenericNew;
-    pygobject_register_class(d, "GObject", G_TYPE_OBJECT,
-			     &PyGObject_Type, NULL);
-    PyDict_SetItemString(PyGObject_Type.tp_dict, "__gdoc__",
-			 pyg_object_descr_doc_get());
-    pyg_set_object_has_new_constructor(G_TYPE_OBJECT);
-
-      /* GObject properties descriptor */
-    if (PyType_Ready(&PyGProps_Type) < 0)
-        return;
-    if (PyType_Ready(&PyGPropsDescr_Type) < 0)
-        return;
-    if (PyType_Ready(&PyGPropsIter_Type) < 0)
-        return;
-    descr = PyObject_New(PyObject, &PyGPropsDescr_Type);
-    PyDict_SetItemString(PyGObject_Type.tp_dict, "props", descr);
-    PyDict_SetItemString(PyGObject_Type.tp_dict, "__module__",
-                        o=PyString_FromString("gobject._gobject"));
-    Py_DECREF(o);
- 
-    REGISTER_GTYPE(d, PyGInterface_Type, "GInterface", G_TYPE_INTERFACE);
-    PyDict_SetItemString(PyGInterface_Type.tp_dict, "__doc__",
-			 pyg_object_descr_doc_get());
-    PyDict_SetItemString(PyGInterface_Type.tp_dict, "__gdoc__",
-			 pyg_object_descr_doc_get());
-
-    REGISTER_GTYPE(d, PyGBoxed_Type, "GBoxed", G_TYPE_BOXED);
-    REGISTER_GTYPE(d, PyGPointer_Type, "GPointer", G_TYPE_POINTER); 
-    PyGEnum_Type.tp_base = &PyInt_Type;
-    REGISTER_GTYPE(d, PyGEnum_Type, "GEnum", G_TYPE_ENUM);
-    PyGFlags_Type.tp_base = &PyInt_Type;
-    REGISTER_GTYPE(d, PyGFlags_Type, "GFlags", G_TYPE_FLAGS);
-
-    PyType_Ready(&PyGObjectWeakRef_Type);
-    PyDict_SetItemString(d, "GObjectWeakRef", (PyObject *) &PyGObjectWeakRef_Type);
-
-    /* pygobject version */
-    tuple = Py_BuildValue ("(iii)",
-			   PYGOBJECT_MAJOR_VERSION,
-			   PYGOBJECT_MINOR_VERSION,
-			   PYGOBJECT_MICRO_VERSION);
-    PyDict_SetItemString(d, "pygobject_version", tuple);
-    /* backwards compatibility */
-    PyDict_SetItemString(d, "pygtk_version", tuple);
-    Py_DECREF(tuple);
-
-    /* for addon libraries ... */
-    PyDict_SetItemString(d, "_PyGObject_API",
-			 o=PyCObject_FromVoidPtr(&pygobject_api_functions,NULL));
-    Py_DECREF(o);
-
-    /* features */
-    features = PyDict_New();
-#ifdef HAVE_FFI_H
-    PyDict_SetItemString(features, "generic-c-marshaller", Py_True);
-#endif
-    PyDict_SetItemString(d, "features", features);
-    Py_DECREF(features);
-
-    /* some constants */
+/* some constants */
+static void
+pygobject_register_constants(PyObject *m)
+{
     PyModule_AddIntConstant(m, "SIGNAL_RUN_FIRST", G_SIGNAL_RUN_FIRST);
     PyModule_AddIntConstant(m, "SIGNAL_RUN_LAST", G_SIGNAL_RUN_LAST);
     PyModule_AddIntConstant(m, "SIGNAL_RUN_CLEANUP", G_SIGNAL_RUN_CLEANUP);
@@ -2795,18 +2540,79 @@ init_gobject(void)
     /* The rest of the types are set in __init__.py */
     PyModule_AddObject(m, "TYPE_INVALID", pyg_type_wrapper_new(G_TYPE_INVALID));
     PyModule_AddObject(m, "TYPE_GSTRING", pyg_type_wrapper_new(G_TYPE_GSTRING));
+}
+
+/* features */
+static void
+pygobject_register_features(PyObject *d)
+{
+    PyObject *features;
     
-    pyg_register_gtype_custom(G_TYPE_STRV,
-			      _pyg_strv_from_gvalue,
-			      _pyg_strv_to_gvalue);
+    features = PyDict_New();
+#ifdef HAVE_FFI_H
+    PyDict_SetItemString(features, "generic-c-marshaller", Py_True);
+#endif
+    PyDict_SetItemString(d, "features", features);
+    Py_DECREF(features);
+}
+
+static void
+pygobject_register_version_tuples(PyObject *d)
+{
+    PyObject *tuple;
+
+    /* pygobject version */
+    tuple = Py_BuildValue ("(iii)",
+			   PYGOBJECT_MAJOR_VERSION,
+			   PYGOBJECT_MINOR_VERSION,
+			   PYGOBJECT_MICRO_VERSION);
+    PyDict_SetItemString(d, "pygobject_version", tuple);
+
+    /* backwards compatibility */
+    PyDict_SetItemString(d, "pygtk_version", tuple);
+    Py_DECREF(tuple);
+}
+
+static void
+pygobject_register_warnings(PyObject *d)
+{
+    PyObject *warning;
 
     warning = PyErr_NewException("gobject.Warning", PyExc_Warning, NULL);
     PyDict_SetItemString(d, "Warning", warning);
     add_warning_redirection("GLib", warning);
     add_warning_redirection("GLib-GObject", warning);
     add_warning_redirection("GThread", warning);
+}
 
+
+PYGLIB_MODULE_START(_gobject, "gobject._gobject")
+{
+    PyObject *d;
+
+    g_type_init();
+    pyglib_init();
+
+    d = PyModule_GetDict(module);
+    pygobject_register_api(d);
+    pygobject_register_constants(module);
+    pygobject_register_features(d);
+    pygobject_register_version_tuples(d);
+    pygobject_register_warnings(d);
+    pygobject_type_register_types(d);
+    pygobject_object_register_types(d);
+    pygobject_interface_register_types(d);
+    pygobject_paramspec_register_types(d);
+    pygobject_boxed_register_types(d);
+    pygobject_pointer_register_types(d);
+    pygobject_enum_register_types(d);
+    pygobject_flags_register_types(d);
+    
       /* signal registration recognizes this special accumulator 'constant' */
     _pyg_signal_accumulator_true_handled_func = \
         PyDict_GetItemString(d, "signal_accumulator_true_handled");
+
+    pygobject_api_functions.threads_enabled = pyglib_threads_enabled();
+    _pyglib_notify_on_enabling_threads(pyg_note_threads_enabled);
 }
+PYGLIB_MODULE_END
