@@ -221,7 +221,7 @@ class TestFile(unittest.TestCase):
             finally:
                 loop.quit()
 
-        self.file.query_info_async(callback, "standard")
+        self.file.query_info_async("standard", callback)
 
         loop = glib.MainLoop()
         loop.run()
@@ -295,6 +295,36 @@ class TestFile(unittest.TestCase):
         finally:
             os.unlink("copy.txt")
 
+    def test_copy_async(self):
+        if os.path.exists('copy.txt'):
+            os.unlink("copy.txt")
+
+        source = gio.File('file.txt')
+        destination = gio.File('copy.txt')
+
+        def copied(source_, result):
+            try:
+                self.assert_(source_ is source)
+                self.failUnless(source_.copy_finish(result))
+            finally:
+                loop.quit()
+
+        def progress(current, total):
+            self.assert_(isinstance(current, long))
+            self.assert_(isinstance(total, long))
+            self.assert_(0 <= current <= total)
+
+        try:
+            loop = glib.MainLoop()
+            source.copy_async(destination, copied, progress_callback = progress)
+            loop.run()
+
+            self.failUnless(os.path.exists('copy.txt'))
+            self.assertEqual(open('file.txt').read(),
+                             open('copy.txt').read())
+        finally:
+            os.unlink("copy.txt")
+
     # See bug 546591.
     def test_copy_progress(self):
         source = gio.File('file.txt')
@@ -351,7 +381,8 @@ class TestFile(unittest.TestCase):
                 self.assertEqual(info.type, gio.FILE_ATTRIBUTE_TYPE_UINT64)
                 self.assertEqual(info.name, "time::modified")
                 self.assertEqual(info.flags,
-                                 gio.FILE_ATTRIBUTE_INFO_COPY_WHEN_MOVED)
+                                 gio.FILE_ATTRIBUTE_INFO_COPY_WHEN_MOVED |
+                                 gio.FILE_ATTRIBUTE_INFO_COPY_WITH_FILE)
 
     def testQueryWritableNamespaces(self):
         infolist = self.file.query_writable_namespaces()
@@ -373,6 +404,24 @@ class TestFile(unittest.TestCase):
                                               10, gio.FILE_QUERY_INFO_NONE)
                 self.assertEqual(ret, True)
 
+    def testSetAttributesAsync(self):
+        def callback(gfile, result):
+            try:
+                info = gfile.set_attributes_finish(result)
+                usec = info.get_attribute_uint32("time::modified-usec")
+                self.assertEqual(usec, 10)
+            finally:
+                loop.quit()        
+
+        info = gio.FileInfo()
+        info.set_attribute_uint32("time::modified-usec", 10)
+        
+        canc = gio.Cancellable()
+        self.file.set_attributes_async(info, callback)
+        
+        loop = glib.MainLoop()
+        loop.run()
+
     def testReplaceContents(self):
         self.file.replace_contents("testing replace_contents")
         cont, leng, etag = self.file.load_contents()
@@ -423,6 +472,22 @@ class TestFile(unittest.TestCase):
         self.assertEquals(hash(gio.File('foo')),
                           hash(gio.File('foo')))
 
+    def testSetDisplayNameAsync(self):
+        def callback(gfile, result):
+            try:
+                new_gfile = gfile.set_display_name_finish(result)
+                new_name = new_gfile.get_basename()
+                self.assertEqual(new_name, "new.txt")
+                deleted = new_gfile.delete()
+                self.assertEqual(deleted, True)
+            finally:
+                loop.quit()        
+
+        canc = gio.Cancellable()
+        self.file.set_display_name_async("new.txt", callback, cancellable=canc)
+        
+        loop = glib.MainLoop()
+        loop.run()
 
 class TestGFileEnumerator(unittest.TestCase):
     def setUp(self):
@@ -470,6 +535,21 @@ class TestGFileEnumerator(unittest.TestCase):
         loop = glib.MainLoop()
         loop.run()
 
+    def testCloseFilesAsync(self):
+        def callback(enumerator, result):
+            try:
+                enumerator.close_finish(result)
+            finally:
+                loop.quit()
+
+        enumerator = self.file.enumerate_children(
+            "standard::*", gio.FILE_QUERY_INFO_NOFOLLOW_SYMLINKS)
+
+        enumerator.close_async(callback)
+
+        loop = glib.MainLoop()
+        loop.run()
+
 
 class TestInputStream(unittest.TestCase):
     def setUp(self):
@@ -500,6 +580,26 @@ class TestInputStream(unittest.TestCase):
                                              50),
                           some_data)
 
+    def testSkip(self):
+        self.stream.skip(2)
+        res = self.stream.read()
+        self.assertEqual(res, "sting")
+        
+    def testSkipAsync(self):
+        def callback(stream, result):
+            try:
+                size = stream.skip_finish(result)
+                self.assertEqual(size, 2)
+                res = stream.read()
+                self.assertEqual(res, "sting")
+            finally:
+                loop.quit()
+        
+        self.stream.skip_async(2, callback)
+
+        loop = glib.MainLoop()
+        loop.run()
+
     def test_read_part(self):
         self.assertEquals(self._read_in_loop(self.stream,
                                              lambda: self.stream.read_part()),
@@ -704,7 +804,38 @@ class TestOutputStream(unittest.TestCase):
 
         loop = glib.MainLoop()
         loop.run()
+        
+    def testFlushAsync(self):
+        def callback(stream, result):
+            try:
+                self.failUnless(stream.flush_finish(result))
+            finally:
+                loop.quit()
 
+        self.stream.flush_async(callback)
+
+        loop = glib.MainLoop()
+        loop.run()
+    
+    def testSpliceAsync(self):
+        _f = open("stream.txt", "w+")
+        _f.write("testing")
+        _f.seek(0)
+        instream = gio.unix.InputStream(_f.fileno(), False)
+        
+        def callback(stream, result):
+            try:
+                size = stream.splice_finish(result)
+                self.assertEqual(size, 7)
+                
+            finally:
+                os.unlink("stream.txt")
+                loop.quit()
+
+        self.stream.splice_async(instream, callback)
+
+        loop = glib.MainLoop()
+        loop.run()
 
 class TestMemoryOutputStream(unittest.TestCase):
     def setUp(self):
@@ -797,3 +928,18 @@ class TestVfs(unittest.TestCase):
     def testGetSupportedURISchemes(self):
         result = self.vfs.get_supported_uri_schemes()
         self.failUnless(type(result), [])
+
+class TestVolume(unittest.TestCase):
+    def setUp(self):
+        self.monitor = gio.volume_monitor_get()
+    
+    def testVolumeEnumerate(self):
+        volumes = self.monitor.get_volumes()
+        self.failUnless(isinstance(volumes, list))
+        for v in volumes:
+            if v is not None:
+                ids = v.enumerate_identifiers()
+                self.failUnless(isinstance(ids, list))
+                for id in ids:
+                    if id is not None:
+                        self.failUnless(isinstance(id, str))
